import logging
import os.path
import httpx
import json
import time
from abc import ABC, abstractmethod
from appdirs import site_data_dir
from os.path import exists
from stream_unzip import stream_unzip
from random import randrange
import pytz
from datetime import datetime, timedelta
from typing import List, Dict
import xml.etree.ElementTree as ET


class IssueTimeCollector:

    def __init__(self):
        self.issue_time_utc = None

    def consume(self, event, elem):
        if elem.tag.endswith("IssueTime"):
            if elem.text is not None:
                self.issue_time_utc = datetime.fromisoformat(elem.text.replace("Z", "+00:00"))


class TimeStepsCollector:

    def __init__(self):
        self.__is_collecting = False
        self.utc_timesteps = []

    def consume(self, event, elem):
        if event == 'start' and elem.tag.endswith("ForecastTimeSteps"):
            self.utc_timesteps.clear()
            self.__is_collecting = True
        elif event == 'start' and self.__is_collecting and elem.tag.endswith("TimeStep"):
            if elem.text is not None:
                utc_time = datetime.fromisoformat(elem.text.replace("Z", "+00:00"))
                self.utc_timesteps.append(utc_time)
        elif self.__is_collecting and event == 'end' and elem.tag.endswith("ForecastTimeSteps"):
            self.__is_collecting = False


class ForecastValuesCollector:

    def __init__(self, station_id: str):
        self.station_id = station_id
        self.__parameter_name = ""
        self.parameters: Dict[str, List[float]] = {}
        self.__is_placemark_collecting = False
        self.__is_forecast_collecting = False

    def consume(self, event, elem):
        if event == 'start' and elem.tag.endswith("Placemark"):
            self.__is_placemark_collecting = True

        elif self.__is_placemark_collecting and elem.tag.endswith("name"):
            if elem.text is not None:
                if self.station_id == elem.text:
                    self.__is_forecast_collecting = True
                    self.parameters.clear()
                else:
                    self.__is_forecast_collecting = False

        elif self.__is_forecast_collecting and event == 'start' and elem.tag.endswith("Forecast"):
            for name in elem.attrib.keys():
                if name.endswith("elementName"):
                    self.__parameter_name = elem.attrib[name]
                    break

        elif self.__is_forecast_collecting and elem.tag.endswith("value"):
            if elem.text is not None:
                values = elem.text.split()
                self.parameters[self.__parameter_name] = [(None if value == '-' else float(value)) for value in values]

        elif event == 'end' and elem.tag.endswith("Placemark"):
            self.__is_placemark_collecting = False


class ParameterUtcSeries:

    TIME_PATTERN = "%Y.%m.%dT%H"

    @staticmethod
    def __datetime_to_utc_string(dt: datetime) -> str:
        return dt.astimezone(pytz.UTC).strftime(ParameterUtcSeries.TIME_PATTERN)

    @staticmethod
    def __utc_string_to_datetime(utc_string: str) -> datetime:
        return datetime.strptime(utc_string, ParameterUtcSeries.TIME_PATTERN)

    @staticmethod
    def create(name: str, time_series: List[datetime], values: Dict[str, List[float]]):
        return ParameterUtcSeries({ParameterUtcSeries.__datetime_to_utc_string(time_series[i]): values[name][i] for i in range(0, len(time_series))})

    def __init__(self, utc_series: Dict[str, float]):
        self.utc_series = utc_series

    def size(self) -> int:
        return len(self.utc_series.keys())

    @property
    def date_from_utc(self) -> datetime:
        return self.__utc_string_to_datetime(sorted(self.utc_series.keys())[0])

    @property
    def date_to_utc(self) -> datetime:
        return self.__utc_string_to_datetime(sorted(self.utc_series.keys())[-1])

    def value_at(self, dt: datetime) -> float:
        return self.utc_series.get(self.__datetime_to_utc_string(dt))

    def merge(self, old, min_datetime: datetime):
        series = {time_utc: old.utc_series[time_utc] for time_utc in old.utc_series.keys() if time_utc >= self.__datetime_to_utc_string(min_datetime)}
        series.update(self.utc_series)
        return ParameterUtcSeries(series)


class MosmixS:

    @staticmethod
    def create(station_id: str,
               issue_time_utc: datetime,
               timesteps_utc: List[datetime],
               parameters: Dict[str, List[float]]):
        return MosmixS(station_id,
                       issue_time_utc,
                       {parameter: ParameterUtcSeries.create(parameter, timesteps_utc, parameters) for parameter in parameters})

    def __utc_to_local(utc: datetime) -> datetime:
        return datetime.strptime((utc + (datetime.now() - datetime.utcnow())).strftime("%d.%m.%Y %H:%M:%S.%f"), "%d.%m.%Y %H:%M:%S.%f")

    def __init__(self,
                 station_id: str,
                 issue_time_utc: datetime,
                 parameter_series: Dict[str, ParameterUtcSeries]):
        self.station_id = station_id
        self.__issue_time_utc = issue_time_utc
        self.__parameter_series = parameter_series
        self.issue_time = MosmixS.__utc_to_local(self.__issue_time_utc)
        self.date_from = MosmixS.__utc_to_local(self.__parameter_series["Rad1h"].date_from_utc)
        self.date_to = MosmixS.__utc_to_local(self.__parameter_series["Rad1h"].date_to_utc)

    def merge(self, old_mosmix):
        if old_mosmix is None:
            return self
        else:
            min_local_datetime = datetime.now() - timedelta(days=5)
            merged = MosmixS(self.station_id,
                             self.__issue_time_utc,
                             {parameter: self.__parameter_series[parameter].merge(old_mosmix.__parameter_series[parameter], min_local_datetime) for parameter in self.__parameter_series.keys()})
            logging.debug("merging \nold mosmix:    " + str(old_mosmix) + " \nnew mosmix:    " + str(self) + " \nmerged mosmix: " + str(merged))
            return merged

    def is_expired(self) -> bool:
        content_age_min = int((datetime.now() - self.issue_time).total_seconds() / 60)
        return content_age_min > (60 + 25 + randrange(5))

    def supports(self, dt: datetime) -> bool:
        return self.date_from <= dt <= self.date_to

    def rad1h(self, dt: datetime) -> float:
        return self.__parameter_series["Rad1h"].value_at(dt)

    def sund1(self, dt: datetime) -> float:
        return self.__parameter_series["SunD1"].value_at(dt)

    def neff(self, dt: datetime) -> float:
        return self.__parameter_series["Neff"].value_at(dt)

    def wwm(self, dt: datetime) -> float:
        return self.__parameter_series["wwM"].value_at(dt)

    def vv(self, dt: datetime) -> float:
        return self.__parameter_series["VV"].value_at(dt)

    def __str__(self):
        return str(self.__parameter_series["Rad1h"].size()) + " entries (issued=" + self.issue_time.strftime("%Y.%m.%dT%H:%M") + "; " + self.date_from.strftime("%Y.%m.%dT%H:%M") + " -> " + self.date_to.strftime("%Y.%m.%dT%H:%M") + ")"

    def save(self, filename: str = "mosmix.json"):
        with open(filename, "w") as file:
            data = json.dumps({ "station_id": self.station_id,
                                "issue_time_utc": self.__issue_time_utc.isoformat(),
                                "parameter_series": { parameter: self.__parameter_series[parameter].utc_series for parameter in self.__parameter_series.keys()}})
            file.write(data)

    @staticmethod
    def load(filename: str = "mosmix.json"):
        if exists(filename):
            with open(filename, "r") as file:
                try:
                    data = json.loads(file.read())
                    station_id = data['station_id']
                    issue_time_utc = datetime.fromisoformat(data['issue_time_utc'])
                    parameter_series = {parameter: ParameterUtcSeries(data['parameter_series'][parameter]) for parameter in data['parameter_series'].keys()}
                    return MosmixS(station_id, issue_time_utc, parameter_series)
                except Exception as e:
                    logging.warning("error occurred loading mosmix cache file " + filename, e)
        return None


class MosmixLoader(ABC):

    @abstractmethod
    def get(self) -> MosmixS:
        pass


class MosmixWebLoader:

    def __init__(self, station_id: str):
        self.__station_id = station_id

    @staticmethod
    def __perform_get_chunked(url: str):
        with httpx.stream('GET', url) as r:
            yield from r.iter_bytes(chunk_size=65536)

    def get(self) -> MosmixS:
        url = 'https://opendata.dwd.de/weather/local_forecasts/mos/MOSMIX_S/all_stations/kml/MOSMIX_S_LATEST_240.kmz'

        issue_time_collector = IssueTimeCollector()
        timesteps_collector = TimeStepsCollector()
        forecasts_collector = ForecastValuesCollector(self.__station_id)

        xml_parser = ET.XMLPullParser(['start', 'end'])
        for file_name, file_size, unzipped_chunks in stream_unzip(MosmixWebLoader.__perform_get_chunked(url)):
            for chunk in unzipped_chunks:
                xml_parser.feed(chunk)
                for event, elem in xml_parser.read_events():
                    issue_time_collector.consume(event, elem)
                    timesteps_collector.consume(event, elem)
                    forecasts_collector.consume(event, elem)
        mosmix = MosmixS.create(self.__station_id,
                                issue_time_collector.issue_time_utc,
                                timesteps_collector.utc_timesteps,
                                forecasts_collector.parameters)
        return mosmix


class FileCachedMosmixLoader(MosmixLoader):

    def __init__(self, station_id: str):
        self.__station_id = station_id
        self.__parent_loader = MosmixWebLoader(self.__station_id)
        self.__dir = site_data_dir("pv_forecast", appauthor=False)
        if not exists(self.__dir):
            os.makedirs(self.__dir)

    def get(self) -> MosmixS:
        cache_filename = os.path.join(self.__dir, "mosmixs_" + self.__station_id + ".json")
        mosmix = MosmixS.load(cache_filename)
        if mosmix is not None:
            if mosmix.is_expired():
                if int((time.time() - os.path.getmtime(cache_filename)) / 60) < 10:   # at maximum all 10 min the (large!) mosmix file will be loaded via web
                    return mosmix
            else:
                return mosmix
        new_mosmix = self.__parent_loader.get()
        merged_mosmix = new_mosmix.merge(mosmix)
        merged_mosmix.save(cache_filename)
        logging.info("Mosmix file updated " + str(merged_mosmix))
        return merged_mosmix


class MemoryCachedMosmixLoader(MosmixLoader):

    def __init__(self, station_id: str):
        self.__parent_loader = FileCachedMosmixLoader(station_id)
        self.__cached_mosmix = None

    def get(self) -> MosmixS:
        if self.__cached_mosmix is None or self.__cached_mosmix.is_expired():
            mosmix = self.__parent_loader.get()
            self.__cached_mosmix = mosmix
        return self.__cached_mosmix





import unittest
import tempfile
import os
import pytz
from random import randrange
from pvpower.mosmix import MosmixS, MemoryCachedMosmixLoader
from datetime import datetime, timedelta



class TestMosmix(unittest.TestCase):

    def test_load_from_web(self):
        mosmix = MemoryCachedMosmixLoader('N0677').get()
        tomorrow = datetime.now() + timedelta(days=1)
        irradiance = mosmix.rad1h(tomorrow)
        self.assertTrue(irradiance >= 0)
        #print("issue time utc:   " + str(mosmix.issue_time_utc))
        #print("issue time local: " + str(mosmix.issue_time))

    def test_save_and_restore(self):
        mosmix = MemoryCachedMosmixLoader('N0677').get()
        with tempfile.TemporaryDirectory() as tmp_dir:
            temp_file = os.path.join(tmp_dir, str(randrange(100000)) + ".temp")
            mosmix.save(temp_file)
            restored_mosmix = MosmixS.load(temp_file)
            self.assertEqual(mosmix.station_id, restored_mosmix.station_id)
            self.assertEqual(mosmix.issue_time, restored_mosmix.issue_time)
            self.assertEqual(mosmix.is_expired(), restored_mosmix.is_expired())
            time = datetime.now() + timedelta(days=5)
            self.assertEqual(mosmix.supports(time), restored_mosmix.supports(time))
            self.assertEqual(mosmix.vv(time), restored_mosmix.vv(time))
            os.remove(temp_file)

    def test_utc_local_time(self):
        mosmix = MemoryCachedMosmixLoader('N0677').get()

        print("cross-check with https://mosmix.de/online.html#/station/N0677/")
        print("time (local) ............. time (utc) .............. vv")
        now_local = datetime.now()
        for i in range(0, 10):
            tomorrow = datetime.strptime((datetime.now() + timedelta(days=1)).strftime("%d.%m.%Y %H") + ":00", "%d.%m.%Y %H:%M") + timedelta(hours=i)
            local = tomorrow.strftime("%d.%m.%Y %H:%M")
            vv = str(mosmix.vv(tomorrow))
            utc = tomorrow.astimezone(pytz.UTC).strftime("%d.%m.%Y %H:%M")
            print(local + " " + "".join(["."] * (25 - len(utc))) + " " + utc + " " + "".join(["."] * (15 - len(vv))) + " " + vv)


if __name__ == '__main__':
    unittest.main()


import logging
import pytz
from dataclasses import dataclass
from datetime import datetime
from abc import ABC, abstractmethod
from sklearn import svm
from typing import List
from pvpower.weather_forecast import WeatherForecast
from pvpower.traindata import LabelledWeatherForecast, TrainData



class Vectorizer(ABC):

    def __init__(self, datetime_resolution_minutes: int = 15):
        self.datetime_resolution_minutes = datetime_resolution_minutes

    def _utc_minutes_of_day(self, dt: datetime) -> int:
        utc_time = dt.astimezone(pytz.UTC)
        minutes = (utc_time.hour * 60) + utc_time.minute
        minutes_rounded = int(minutes / self.datetime_resolution_minutes) * self.datetime_resolution_minutes
        return minutes_rounded

    def _scale(self, value: int, max_value: int, digits=1) -> float:
        if value == 0:
            return 0
        else:
            return round(value * 100 / max_value, digits)

    @abstractmethod
    def vectorize(self, sample: WeatherForecast) -> List[float]:
        pass



class CoreVectorizer(Vectorizer):

    def vectorize(self, sample: WeatherForecast) -> List[float]:
        vectorized = [self._scale(sample.time_utc.month, 12),
                      self._scale(self._utc_minutes_of_day(sample.time_utc), 24 * 60),
                      self._scale(sample.irradiance, 1000)]
        return vectorized

    def __str__(self):
        return "CoreVectorizer"


class SunshineVectorizer(Vectorizer):

    def vectorize(self, sample: WeatherForecast) -> List[float]:
        vectorized = [self._scale(sample.time_utc.month, 12),
                      self._scale(self._utc_minutes_of_day(sample.time_utc), 24 * 60),
                      self._scale(sample.sunshine, 5000)]
        return vectorized

    def __str__(self):
        return "SunshineVectorizer"


class SushinePlusCloudCoverVectorizer(SunshineVectorizer):

    def vectorize(self, sample: WeatherForecast) -> List[float]:
        return super().vectorize(sample) +  [self._scale(sample.cloud_cover, 200)]

    def __str__(self):
        return "Sunshine+CloudVectorizer"



class PlusVisibilityVectorizer(CoreVectorizer):

    def vectorize(self, sample: WeatherForecast) -> List[float]:
        return super().vectorize(sample) +  [self._scale(sample.visibility, 50000)]

    def __str__(self):
        return "Core+VisibilityVectorizer"


class PlusSunshineVectorizer(CoreVectorizer):

    def vectorize(self, sample: WeatherForecast) -> List[float]:
        return super().vectorize(sample) +  [self._scale(sample.sunshine, 5000)]

    def __str__(self):
        return "Core+SunshineVectorizer"


class PlusCloudCoverVectorizer(CoreVectorizer):

    def vectorize(self, sample: WeatherForecast) -> List[float]:
        return super().vectorize(sample) +  [self._scale(sample.cloud_cover, 200)]

    def __str__(self):
        return "Core+CloudVectorizer"


class PlusVisibilitySunshineVectorizer(CoreVectorizer):

    def vectorize(self, sample: WeatherForecast) -> List[float]:
        return super().vectorize(sample) +  [self._scale(sample.visibility, 50000),
                                             self._scale(sample.sunshine, 5000)]

    def __str__(self):
        return "Core+Visibility+SunshineVectorizer"


class PlusVisibilityCloudCoverVectorizer(CoreVectorizer):

    def vectorize(self, sample: WeatherForecast) -> List[float]:
        return super().vectorize(sample) +  [self._scale(sample.visibility, 50000),
                                             self._scale(sample.cloud_cover, 200)]

    def __str__(self):
        return "Core+Visibility+CloudVectorizer"


class PlusVisibilityFogCloudCoverVectorizer(CoreVectorizer):

    def vectorize(self, sample: WeatherForecast) -> List[float]:
        return super().vectorize(sample) +  [self._scale(sample.visibility, 50000),
                                             self._scale(sample.cloud_cover, 200),
                                             self._scale(sample.probability_for_fog, 100)]

    def __str__(self):
        return "Core+Visibility+Fog+CloudVectorizer"


class FullVectorizer(CoreVectorizer):

    def vectorize(self, sample: WeatherForecast) -> List[float]:
        return super().vectorize(sample) +  [self._scale(sample.visibility, 50000),
                                             self._scale(sample.cloud_cover, 200),
                                             self._scale(sample.probability_for_fog, 100),
                                             self._scale(sample.sunshine, 5000)]

    def __str__(self):
        return "FullVectorizer"


@dataclass(frozen=True)
class TrainReport:
    samples: List[LabelledWeatherForecast]


class TestReport:

    def __init__(self, train_report: TrainReport, validation_samples: List[LabelledWeatherForecast], predictions: List[int]):
        self.train_report = train_report
        self.validation_samples = validation_samples
        self.predictions = predictions

    def __percent(self, real, predicted):
        if real == 0:
            return 0
        elif predicted == 0:
            return 0
        else:
            return round((predicted * 100 / real) - 100, 2)

    def __diff_all(self) -> List[float]:
        diff_total = []
        for i in range(len(self.validation_samples)):
            if self.validation_samples[i].irradiance == 0:
                continue
            else:
                predicted = self.predictions[i]

            real = self.validation_samples[i].power_watt
            if real == 0:
                if predicted == 0:     # ignore true 0 predictions to void wasting the score
                    diff = 0
                else:
                    diff = 100
            else:
                if real < 10 and abs(predicted - real) < 10:  # ignore diff < 10
                    diff = 0
                else:
                    diff = self.__percent(real, predicted)
            diff_total.append(diff)
        return sorted(diff_total)

    @property
    def score(self) -> float:
        values = sorted(list(self.__diff_all()))
        values = values[2:-2]
        abs_values = [abs(value) for value in values]
        if abs_values == 0:
            return 10000
        else:
            return round(sum(abs_values) / len(abs_values), 2)

    def __repr__(self):
        return self.__str__()

    def __str__(self):
        txt = "score:  " + str(self.score) + " (samples: " + str(len(self.train_report.samples)) + ")\n"
        txt += '{:14s}        {:14s} {:14s} {:14s} {:14s} {:14s}         {:10s} {:10s}           {:10s}\n'.format("time", "irradiance", "sunshine", "cloud_cover", "visibility", "proba.fog", "real", "predicted", "diff[%]")
        for i in range(0, len(self.validation_samples)):
            txt += '{:<14s}        {:<14d} {:<14d} {:<14d} {:<14d}  {:<14d}        {:<10d} {:<10d}           {:<10d}\n'.format(self.validation_samples[i].time.strftime("%d.%b  %H:%M"),
                                                                                                                               self.validation_samples[i].irradiance,
                                                                                                                               self.validation_samples[i].sunshine,
                                                                                                                               self.validation_samples[i].cloud_cover,
                                                                                                                               self.validation_samples[i].visibility,
                                                                                                                               self.validation_samples[i].probability_for_fog,
                                                                                                                               self.validation_samples[i].power_watt,
                                                                                                                               self.predictions[i],
                                                                                                                               int(self.__percent(self.validation_samples[i].power_watt, self.predictions[i])))
        txt = txt + "\nscore:  " + str(self.score)
        return txt


class Estimator(ABC):


    @property
    @abstractmethod
    def datetime_resolution_minutes(self) -> int:
        pass

    @abstractmethod
    def date_last_train(self) -> datetime:
        pass

    @abstractmethod
    def retrain(self, train_data: TrainData) -> TrainReport:
        pass

    @abstractmethod
    def predict(self, sample: WeatherForecast) -> int:
        pass


class SMVEstimator(Estimator):

    def __init__(self, vectorizer: Vectorizer):
        self.__clf = svm.SVC(kernel='poly') # it seems that the SVM approach produces good predictions. refer https://www.sciencedirect.com/science/article/pii/S136403212200274X?via%3Dihub and https://citeseerx.ist.psu.edu/viewdoc/download?doi=10.1.1.221.4021&rep=rep1&type=pdf
        self.__vectorizer = vectorizer
        self.__date_last_train = datetime.fromtimestamp(0)
        self.__num_samples_last_train = 0
        self.__num_covered_days_last_train = 0
        self.__score = None

    @property
    def datetime_resolution_minutes(self) -> int:
        return self.__vectorizer.datetime_resolution_minutes

    def date_last_train(self) -> datetime:
        return self.__date_last_train

    def retrain(self, train_data: TrainData) -> TrainReport:
        samples = train_data.samples
        feature_vector_list = [self.__vectorizer.vectorize(sample) for sample in samples]
        label_list = [sample.power_watt for sample in samples]
        if len(set(label_list)) > 1:
            self.__clf.fit(feature_vector_list, label_list)
            self.__date_last_train = datetime.now()
            self.__num_samples_last_train = len(samples)
            self.__num_covered_days_last_train = len(set([sample.time.strftime("%Y.%m.%d") for sample in samples]))
            logging.debug("estimator has been (re)trained " + str(self))
        return TrainReport(samples)

    def predict(self, sample: WeatherForecast) -> int:
        if sample.irradiance > 0:
            if self.__num_samples_last_train < 1:
                logging.warning("estimator has not been trained (insufficient train data available). returning 0")
                return 0
            else:
                feature_vector = self.__vectorizer.vectorize(sample)
                predicted = int(self.__clf.predict([feature_vector])[0])
                logging.debug(str(predicted) + " watt predicted for " + str(sample) + " (features: " + str(feature_vector) + ")")
                if predicted >= 0:
                    return predicted
                else:
                    logging.debug("predicted value is " + str(predicted) + " correct them to 0")
                    return 0
        else:
            return 0

    def __str__(self):
        return "SVMEstimator(vectorizer=" + str(self.__vectorizer) + "; deviation: " + ("unknown" if self.__score is None else str(round(self.__score, 1)) +"%") + "; trained with " + str(self.__num_samples_last_train) + " samples; age " + str(datetime.now() - self.date_last_train()) + "; time range: " + str(self.__num_covered_days_last_train) + " days; datetime_resolution: " + str(self.datetime_resolution_minutes) + " min)"




import logging
import os
import pytz
import shutil
import gzip
import tempfile
from time import sleep
from threading import Thread
from pathlib import Path
from os.path import exists
from threading import RLock
from datetime import datetime, timedelta
from typing import List, Optional
from pvpower.weather_forecast import WeatherForecast



class LabelledWeatherForecast(WeatherForecast):

    def __init__(self,
                 time: datetime,
                 irradiance: int,
                 sunshine: int,
                 cloud_cover: int,
                 probability_for_fog: int,
                 visibility: int,
                 power_watt: int):
        super().__init__(time, irradiance, sunshine, cloud_cover, probability_for_fog, visibility)
        self.power_watt = power_watt


    @staticmethod
    def create(weather_forecast: WeatherForecast, power_watt: int, time: datetime = None):

        return LabelledWeatherForecast(weather_forecast.time_utc if time is None else time,
                                       weather_forecast.irradiance,
                                       weather_forecast.sunshine,
                                       weather_forecast.cloud_cover,
                                       weather_forecast.probability_for_fog,
                                       weather_forecast.visibility,
                                       power_watt)

    @staticmethod
    def __to_string(value):
        return "" if value is None else str(value)

    def to_csv(self) -> str:
        utc_time = self.time_utc
        return utc_time.strftime("%d.%m.%Y %H:%M") + ";" + \
               LabelledWeatherForecast.__to_string(self.power_watt) + ";" + \
               LabelledWeatherForecast.__to_string(self.irradiance) + ";" + \
               LabelledWeatherForecast.__to_string(self.sunshine) + ";" + \
               LabelledWeatherForecast.__to_string(self.cloud_cover) + ";" + \
               LabelledWeatherForecast.__to_string(self.probability_for_fog) + ";" + \
               LabelledWeatherForecast.__to_string(self.visibility)

    @staticmethod
    def csv_header() -> str:
        return "utc_time;real_pv_power;irradiance;sunshine;cloud_cover;probability_for_fog;visibility"

    @staticmethod
    def __to_int(txt) -> Optional[int]:
        if len(txt) > 0:
            return int(float(txt))
        else:
            return None

    @staticmethod
    def from_csv(line: str):
        parts = line.split(";")
        utc_time = datetime.strptime(parts[0] + ":00+00:00", "%d.%m.%Y %H:%M:%S%z")
        real_pv_power = LabelledWeatherForecast.__to_int(parts[1])
        irradiance = LabelledWeatherForecast.__to_int(parts[2])
        sunshine = LabelledWeatherForecast.__to_int(parts[3])
        cloud_cover_effective = LabelledWeatherForecast.__to_int(parts[4])
        probability_for_fog = LabelledWeatherForecast.__to_int(parts[5])
        visibility = LabelledWeatherForecast.__to_int(parts[6])
        sample = LabelledWeatherForecast(utc_time, irradiance, sunshine, cloud_cover_effective, probability_for_fog, visibility, real_pv_power)
        return sample


class TrainData:

    def __init__(self, samples: List[LabelledWeatherForecast]):
        self.samples = self.__clean_data(samples)

    def __iter__(self):
        return iter(self.samples)

    def __len__(self):
        return len(self.samples)

    @staticmethod
    def __clean_data(samples: List[LabelledWeatherForecast]) -> List[LabelledWeatherForecast]:
        seen = list()
        samples = list(filter(lambda sample: seen.append(sample.time_utc) is None if sample.time_utc not in seen else False, samples))
        samples = [sample for sample in samples if sample.irradiance > 0]
        return samples

    @staticmethod
    def __rotate(samples: List[LabelledWeatherForecast], offset_percent: int) -> List[LabelledWeatherForecast]:
        step = int(offset_percent * len(samples) / 100)
        return samples[step:] + samples[:step]

    def split(self, ratio_percent:int = 67, offset_percent: int = 0):
        rotated_samples = self.__rotate(self.samples, offset_percent)
        size = len(rotated_samples)
        left = []
        right = []
        for i in range(0, size):
            if i > round(size * ratio_percent / 100, 0):
                right.append(rotated_samples[i])
            else:
                left.append(rotated_samples[i])
        return TrainData(left), TrainData(right)

    def __str__(self):
        return "train data size " + str(len(self.samples)) + " (" + self.samples[0].time.strftime("%Y.%m.%dT%H:%M") + " ... " + self.samples[-1].time.strftime("%Y.%m.%dT%H:%M") + ")"


class TrainSampleLog:
    COMPACTION_PERIOD_DAYS = 15

    def __init__(self, dirname: str):
        self.lock = RLock()
        self.__dirname = dirname
        self.__last_compaction_time = datetime.now() - timedelta(days=self.COMPACTION_PERIOD_DAYS*2)

    def filename(self, compressed: bool = False):
        if compressed:
            fn = os.path.join(self.__dirname, "train.csv.gz")
        else:
            fn = os.path.join(self.__dirname, "train.csv")
        if not exists(fn):
            directory = Path(fn).parent
            if not exists(directory):
                os.makedirs(directory)
        return fn

    def append(self, sample: LabelledWeatherForecast):
        with self.lock:
            # compressed file
            compr_fn = self.filename(compressed=True)
            with gzip.open(compr_fn, "ab") as file:
                line = sample.to_csv() + "\n"
                file.write(line.encode(encoding='UTF-8'))
            logging.debug("train record appended " + str(sample))

        if datetime.now() > (self.__last_compaction_time + timedelta(days=self.COMPACTION_PERIOD_DAYS)):
            self.__last_compaction_time = datetime.now()
            Thread(target=self.compact, args=(15,), daemon=True).start()

    def all(self) -> TrainData:
        with self.lock:
            # plain file
            fn = self.filename()
            if exists(fn):
                try:
                    with open(fn, "rb") as file:
                        lines = [raw_line.decode('UTF-8').strip() for raw_line in file.readlines()]
                        samples = []
                        for line in lines:
                            try:
                                samples.append(LabelledWeatherForecast.from_csv(line))
                            except Exception as e:
                                logging.warning(e)
                        return TrainData(samples)
                except Exception as e:
                    logging.warning("error occurred loading " + fn + " " + str(e))

            # compressed file
            compr_fn = self.filename(compressed=True)
            if exists(compr_fn):
                try:
                    with gzip.open(compr_fn, "rb") as file:
                        lines = [raw_line.decode('UTF-8').strip() for raw_line in file.readlines()]
                        samples = []
                        for line in lines:
                            try:
                                samples.append(LabelledWeatherForecast.from_csv(line))
                            except Exception as e:
                                pass
                        return TrainData(samples)
                except Exception as e:
                    logging.warning("error occurred loading " + compr_fn + " " + str(e))
        return TrainData([])

    def compact(self, delay_sec:int = 0):
        sleep(delay_sec)

        fn = self.filename(compressed=True)
        train_data = self.all()
        min_datetime = datetime.now() - timedelta(days=400)

        with tempfile.TemporaryDirectory() as tmp_dir:
            temp_file = os.path.join(tmp_dir, 'traindata.csv')
            with gzip.open(temp_file, "wb") as file:
                file.write((LabelledWeatherForecast.csv_header() + "\n").encode(encoding='UTF-8'))
                num_samples = len(train_data)
                num_survived = 0
                previous_sample = None
                for sample in train_data:
                    expired = sample.time_utc < min_datetime.astimezone(pytz.UTC)
                    duplicate = False
                    if previous_sample is not None:
                        duplicate = previous_sample.time_utc == sample.time_utc
                    previous_sample = sample
                    if not expired and not duplicate:
                        num_survived += 1
                        line = sample.to_csv() + "\n"
                        file.write(line.encode(encoding='UTF-8'))
            shutil.move(temp_file, fn)
            logging.info("train file " + fn + " compacted (" + str(num_samples - num_survived) + " of " + str(num_samples) + " removed)")

    def __str__(self):
        return "\n".join([sample.to_csv() for sample in self.all()])






import logging
import pickle
from os import path, makedirs
from appdirs import site_data_dir, user_cache_dir
from threading import Thread
from datetime import datetime, timedelta
from typing import Optional
from pvpower.weather_forecast import WeatherStation, WeatherForecast
from pvpower.traindata import LabelledWeatherForecast, TrainSampleLog, TrainData
from pvpower.estimator import Estimator, SMVEstimator, TrainReport, CoreVectorizer


class ValueRecorder:

    def __init__(self, window_size_min: int):
        ts = int(datetime.now().timestamp() / (window_size_min*60)) * (window_size_min*60)
        self.start_time = datetime.fromtimestamp(ts)
        self.end_time = self.start_time + timedelta(minutes=window_size_min)
        self.__power_values = []
        logging.info("value recorder created " + self.start_time.strftime("%Y.%m.%d %H:%M") + " -> " + self.end_time.strftime("%Y.%m.%d %H:%M"))

    def empty(self) -> bool:
        return len(self.__power_values) == 0

    def is_expired(self):
        return datetime.now() > self.end_time

    def add(self, value: int):
        self.__power_values.append(value)

    @property
    def average(self) -> Optional[int]:
        if len(self.__power_values) == 0:
            return None
        else:
            return int(sum(self.__power_values) / len(self.__power_values))

    def __str__(self):
        return self.start_time.strftime("%Y.%m.%d %H:%M") + " -> " + self.end_time.strftime("%Y.%m.%d %H:%M") + "  average power: " + str(self.average) + " num probes: " + str(len(self.__power_values))



class DefaultEstimator(Estimator):

    def __init__(self, estimator: Estimator, pv_forecast_dir: str):
        self.__estimator = estimator
        self.__pv_forecast_dir = pv_forecast_dir

    @property
    def datetime_resolution_minutes(self) -> int:
        return self.__estimator.datetime_resolution_minutes

    @staticmethod
    def __filename() -> str:
        dir = user_cache_dir("pv_forecast", appauthor=False)
        if not path.exists(dir):
            makedirs(dir)
        return path.join(dir, 'dflt_estimator.pickle')

    def date_last_train(self) -> datetime:
        return self.__estimator.date_last_train()

    def predict(self, sample: WeatherForecast) -> int:
        return self.__estimator.predict(sample)

    def retrain(self, train_data: TrainData) -> TrainReport:
        train_report = self.__estimator.retrain(train_data)
        self.__store()
        return train_report

    def __store(self):
        try:
            with open(self.__filename(), 'wb') as file:
                pickle.dump(self.__estimator, file)
        except Exception as e:
            pass

    @staticmethod
    def get(pv_forecast_dir: str, datetime_resolution_minutes: int):
        try:
            with open(DefaultEstimator.__filename(), 'rb') as file:
                estimator = DefaultEstimator(pickle.load(file), pv_forecast_dir)
                if estimator.datetime_resolution_minutes == datetime_resolution_minutes:
                    estimator.retrain(TrainSampleLog(pv_forecast_dir).all())
                    logging.debug("default estimator " + str(estimator) + " loaded from pickle file (" + DefaultEstimator.__filename() + ") and retrained")
                    return estimator
                else:
                    logging.info("datetime_resolution_minutes has been changed")
        except Exception as e:
            logging.warning("error occurred loading estimator " + str(e))

        estimator = DefaultEstimator(SMVEstimator(CoreVectorizer(datetime_resolution_minutes)), pv_forecast_dir)
        estimator.retrain(TrainSampleLog(pv_forecast_dir).all())
        logging.debug("default estimator " + str(estimator) + " created and trained")
        return estimator

    def __str__(self):
        return str(self.__estimator)


class PvPowerForecast:

    def __init__(self, station_id: str, pv_forecast_dir: str = None, estimator: Estimator = None, datetime_resolution_minutes: int = 15):
        self.__datetime_resolution_minutes = datetime_resolution_minutes
        if pv_forecast_dir is None:
            pv_forecast_dir = site_data_dir("pv_forecast", appauthor=False)
        self.train_log = TrainSampleLog(pv_forecast_dir)
        self.weather_forecast_service = WeatherStation(station_id)
        if estimator is None:
            self.__estimator = DefaultEstimator.get(pv_forecast_dir, self.__datetime_resolution_minutes)
        else:
            self.__estimator = estimator
        self.__train_value_recorder = ValueRecorder(self.__estimator.datetime_resolution_minutes)
        self.__date_last_retrain = self.__estimator.date_last_train()
        self.__train_if_old(False)

    def add_current_power_reading(self, real_power: int):
        if self.__train_value_recorder.is_expired():
            try:
                if not self.__train_value_recorder.empty():
                    weather_sample = self.weather_forecast_service.forecast(self.__train_value_recorder.start_time)
                    if weather_sample is not None:
                        annotated_sample = LabelledWeatherForecast.create(weather_sample,
                                                                          self.__train_value_recorder.average,
                                                                          time=self.__train_value_recorder.start_time)
                        self.train_log.append(annotated_sample)
            finally:
                self.__train_value_recorder = ValueRecorder(self.__datetime_resolution_minutes)
        self.__train_value_recorder.add(real_power)

    def predict(self, time: datetime) -> Optional[int]:
        sample = self.weather_forecast_service.forecast(time)
        if sample is None:
            logging.info("could not predict power. Reason: no weather forcast data available (requested date time: " + time.strftime("%Y.%m.%d %H:%M") +
                         ". Available date time range: " + self.weather_forecast_service.forcast_from().strftime("%Y.%m.%d %H:%M") +
                         " -> " + self.weather_forecast_service.forcast_to().strftime("%Y.%m.%d %H:%M")  + "). Returning None")
            return None
        else:
            return self.predict_by_weather_forecast(sample)

    def predict_by_weather_forecast(self, sample: WeatherForecast) -> int:
        try:
            return self.__estimator.predict(sample)
        finally:
            self.__train_if_old()

    def __train_if_old(self, background: bool = True):
        if datetime.now() > (self.__date_last_retrain + timedelta(minutes=3*60)):
            self.__date_last_retrain = datetime.now()
            if background:
                Thread(target=self.__estimator.retrain, daemon=True, args=(self.train_log.all(),)).start()
            else:
                self.__estimator.retrain(self.train_log.all())

    def __str__(self):
        return str(self.__estimator)
